@@ -9,17 +9,21 @@ import os
 import re
 import subprocess
 import sys
+from collections.abc import Generator, Iterator
 from dataclasses import InitVar, dataclass, field
 from datetime import date, datetime
+from functools import lru_cache
 from pathlib import Path
-from typing import Dict, Generator, Iterator, List, Optional, Tuple
+from typing import NoReturn, Optional
 
 import click
 
+from .emoji import demojize, get_emojis
 from .settings import GitConf
 from .utils import (
     Level,
     Profile,
+    load_config,
     make_color,
 )
 
@@ -30,7 +34,9 @@ def load_profile() -> Profile:
     """Load Profile function that return name and email."""
     from .utils import load_pyproject
 
-    _authors: Dict[str, str] = load_pyproject().get("authors", {})
+    _authors: dict[str, str] = (
+        load_pyproject().get("project", {}).get("authors", {})
+    )
     return Profile(
         name=_authors.get(
             "name",
@@ -45,6 +51,9 @@ def load_profile() -> Profile:
         email=_authors.get(
             "email",
             (
+                # FIXME: subprocess.CalledProcessError:
+                #   Command '['git', 'config', '--local', 'user.email']'
+                #   returned non-zero exit status 1.
                 subprocess.check_output(
                     ["git", "config", "--local", "user.email"]
                 )
@@ -55,6 +64,76 @@ def load_profile() -> Profile:
     )
 
 
+@dataclass(frozen=True)
+class CommitPrefix:
+    name: str
+    group: str
+    emoji: str
+
+    def __hash__(self):
+        return hash(self.__str__())
+
+    def __str__(self):
+        return self.name
+
+
+@dataclass(frozen=True)
+class CommitPrefixGroup:
+    name: str
+    emoji: str
+
+    def __hash__(self):
+        return hash(self.__str__())
+
+    def __str__(self):
+        return self.name
+
+
+def get_commit_prefix() -> tuple[CommitPrefix, ...]:
+    """Return tuple of CommitPrefix"""
+    conf: list[str] = load_config().get("git", {}).get("commit_prefix", [])
+    prefix_conf: tuple[str, ...] = tuple(_[0] for _ in conf)
+    return tuple(
+        CommitPrefix(name=n, group=g, emoji=e)
+        for n, g, e in (
+            *conf,
+            *[p for p in GitConf.commit_prefix if p[0] not in prefix_conf],
+        )
+    )
+
+
+def get_commit_prefix_group() -> tuple[CommitPrefixGroup, ...]:
+    """Return tuple of CommitPrefixGroup"""
+    conf: list[str] = (
+        load_config().get("git", {}).get("commit_prefix_group", [])
+    )
+    prefix_conf: tuple[str, ...] = tuple(_[0] for _ in conf)
+    return tuple(
+        CommitPrefixGroup(name=n, emoji=e)
+        for n, e in (
+            *conf,
+            *[
+                p
+                for p in GitConf.commit_prefix_group
+                if p[0] not in prefix_conf
+            ],
+        )
+    )
+
+
+@lru_cache
+def get_git_emojis():
+    cm_prefix = tuple(p.emoji.strip(":") for p in get_commit_prefix())
+    return [emojis for emojis in get_emojis() if emojis["alias"] in cm_prefix]
+
+
+def git_demojize(msg: str):
+    for emojis in get_git_emojis():
+        if (emoji := emojis["emoji"]) in msg:
+            msg = msg.replace(emoji, f':{emojis["alias"]}:')
+    return msg
+
+
 @dataclass
 class CommitMsg:
     """Commit Message dataclass that prepare un-emoji-prefix in that message."""
@@ -66,16 +145,16 @@ class CommitMsg:
     def __str__(self):
         return f"{self.mtype}: {self.content}"
 
-    def __post_init__(self, content: str, mtype: str):
-        self.content: str = self.__prepare_msg(content)
-        if not mtype:
+    def __post_init__(self, content: str, mtype: Optional[str] = None) -> None:
+        self.content: str = self.__prepare_msg(git_demojize(content))
+        if mtype is None:  # pragma: no cover.
             self.mtype: str = self.__gen_msg_type()
 
     def __gen_msg_type(self) -> str:
         if s := re.search(r"^(?P<emoji>:\w+:)\s(?P<prefix>\w+):", self.content):
             prefix: str = s.groupdict()["prefix"]
             return next(
-                (cp[1] for cp in GitConf.commit_prefix if prefix == cp[0]),
+                (cp.group for cp in get_commit_prefix() if prefix == cp.name),
                 "Code Changes",
             )
         return "Code Changes"
@@ -84,12 +163,12 @@ class CommitMsg:
     def mtype_icon(self):
         return next(
             (
-                cpt[1]
-                for cpt in GitConf.commit_prefix_group
-                if cpt[0] == self.mtype
+                cpt.emoji
+                for cpt in get_commit_prefix_group()
+                if cpt.name == self.mtype
             ),
             ":black_nib:",  # ✒️
-        )
+        )  # pragma: no cover
 
     @staticmethod
     def __prepare_msg(content: str) -> str:
@@ -101,16 +180,24 @@ class CommitMsg:
             if ":" in content
             else ("refactored", content)
         )
-        icon: Optional[str] = None
-        for cp in GitConf.commit_prefix:
-            if prefix == cp[0]:
-                icon = f"{cp[2]} "
-        if icon is None:
+        emoji: Optional[str] = None
+        for cp in get_commit_prefix():
+            if prefix == cp.name:
+                emoji = f"{cp.emoji} "
+        if emoji is None:
+            if (
+                load_config()
+                .get("git", {})
+                .get("commit_prefix_force_fix", False)
+            ):
+                _prefix: str = demojize(prefix)
+                if re.match(r"^(?P<emoji>:\w+:)", _prefix):
+                    return f"{_prefix}: {content}"
             raise ValueError(
                 f"The prefix of this commit message does not support, "
                 f"{prefix!r}."
             )
-        return f"{icon}{prefix}: {content.strip()}"
+        return f"{emoji or ''}{prefix}: {content.strip()}"
 
 
 @dataclass(frozen=True)
@@ -131,13 +218,14 @@ class CommitLog:
                 self.msg.content,
                 self.author.name,
                 self.author.email,
+                self.refs,
             )
         )
 
 
-def __validate_for_warning(
-    lines: List[str],
-) -> List[str]:
+def _validate_for_warning(
+    lines: list[str],
+) -> list[str]:
     """Validate Commit message that should to fixed, but it does not impact to
     target repository.
 
@@ -148,7 +236,7 @@ def __validate_for_warning(
     :return: A list of warning message.
     """
     subject: str = lines[0]
-    rs: List[str] = []
+    rs: list[str] = []
 
     # RULE 02: Limit the subject line to 50 characters
     if len(subject) <= 20 or len(subject) > 50:
@@ -171,12 +259,15 @@ def __validate_for_warning(
 
 
 def validate_commit_msg(
-    lines: List[str],
-) -> Tuple[List[str], Level]:
+    lines: list[str],
+) -> tuple[list[str], Level]:
     """Validate Commit message
 
     :param lines: A list of line from commit message.
     :type lines: List[str]
+
+    :rtype: Tuple[List[str], Level]
+    :return: A pair of warning messages and its logging level.
     """
     if not lines:
         return (
@@ -184,9 +275,7 @@ def validate_commit_msg(
             Level.ERROR,
         )
 
-    rs: List[str] = __validate_for_warning(lines)
-
-    has_story_id: bool = False
+    rs: list[str] = _validate_for_warning(lines)
     for line, msg in enumerate(lines[1:], start=2):
         # RULE 06: Wrap the body at 72 characters
         if len(msg) > 72:
@@ -194,12 +283,6 @@ def validate_commit_msg(
                 f"The commit body should wrap at 72 characters at line: {line}."
             )
 
-        if not msg.startswith("["):
-            has_story_id = True
-
-    if not has_story_id and len(lines) > 1:
-        rs.append("Please add a Story ID in the commit message.")
-
     if not rs:
         return (
             ["The commit message has the required pattern."],
@@ -217,7 +300,7 @@ def get_branch_name() -> str:
     )
 
 
-def get_latest_tag(default: bool = True) -> Optional[str]:
+def get_latest_tag(default: bool = True) -> str:
     """Return the latest tag if it exists, otherwise it will return the
     version from ``.__about__`` file.
     """
@@ -230,17 +313,21 @@ def get_latest_tag(default: bool = True) -> Optional[str]:
             .decode(sys.stdout.encoding)
             .strip()
         )
-    except subprocess.CalledProcessError:
+    except subprocess.CalledProcessError as err:
         if default:
             from .__about__ import __version__
 
             return f"v{__version__}"
-        return None
+        raise RuntimeError(
+            "Can not extract the latest version from this project"
+        ) from err
 
 
-def gen_commit_logs(tag2head: str) -> Generator[List[str], None, None]:
+def gen_commit_logs(
+    tag2head: str,
+) -> Generator[list[str], None, None]:  # pragma: no cover.
     """Prepare contents logs to List of commit log."""
-    prepare: List[str] = []
+    prepare: list[str] = []
     for line in (
         subprocess.check_output(
             [
@@ -266,12 +353,14 @@ def get_commit_logs(
     tag: Optional[str] = None,
     *,
     all_logs: bool = False,
-    excluded: Optional[List[str]] = None,
-) -> Iterator[CommitLog]:
+    excluded: Optional[list[str]] = None,
+    is_dt: bool = False,
+) -> Iterator[CommitLog]:  # pragma: no cover.
     """Return a list of commit message logs."""
     from .settings import BumpVerConf
 
-    _exc: List[str] = excluded or [r"^Merge"]
+    regex: str = BumpVerConf.get_regex(is_dt)
+    _exc: list[str] = excluded or [r"^Merge"]
     if tag:
         tag2head: str = f"{tag}..HEAD"
     elif all_logs or not (tag := get_latest_tag(default=False)):
@@ -282,15 +371,13 @@ def get_commit_logs(
     for logs in gen_commit_logs(tag2head):
         if any((re.search(s, logs[1]) is not None) for s in _exc):
             continue
-        header: List[str] = logs[0].split("|")
+        header: list[str] = logs[0].split("|")
         if ref_tag := [
             ref.strip()
             for ref in header[1].strip().split(",")
             if "tag: " in ref
         ]:
-            if search := re.search(
-                rf"tag:\sv(?P<version>{BumpVerConf.regex})", ref_tag[0]
-            ):
+            if search := re.search(rf"tag:\sv(?P<version>{regex})", ref_tag[0]):
                 refs = search.groupdict()["version"]
         yield CommitLog(
             hash=header[0],
@@ -307,7 +394,7 @@ def get_commit_logs(
         )
 
 
-def merge2latest_commit(no_verify: bool = False):
+def merge2latest_commit(no_verify: bool = False):  # pragma: no cover.
     subprocess.run(
         ["git", "commit", "--amend", "--no-edit", "-a"]
         + (["--no-verify"] if no_verify else [])
@@ -318,7 +405,7 @@ def get_latest_commit(
     file: Optional[str] = None,
     edit: bool = False,
     output_file: bool = False,
-) -> List[str]:
+) -> list[str]:  # pragma: no cover.
     if file:
         with Path(file).open(encoding="utf-8") as f_msg:
             raw_msg = f_msg.read().splitlines()
@@ -331,7 +418,7 @@ def get_latest_commit(
             .strip()
             .splitlines()
         )
-    lines: List[str] = [
+    lines: list[str] = [
         msg for msg in raw_msg if not msg.strip().startswith("#")
     ]
     if lines[-1] != "":
@@ -377,31 +464,43 @@ def tg():
 @cli_git.command()
 @click.option("-t", "--tag", type=click.STRING, default=None)
 @click.option("-a", "--all-logs", is_flag=True)
-def log(tag: Optional[str], all_logs: bool):
+@click.option("-d", "--datetime-mode", is_flag=True)
+def log(
+    tag: Optional[str],
+    all_logs: bool,
+    datetime_mode: bool,
+):  # pragma: no cover.
     """Show the Commit Logs from the latest Tag to HEAD."""
     click.echo(
-        "\n".join(str(x) for x in get_commit_logs(tag=tag, all_logs=all_logs)),
+        "\n".join(
+            str(x)
+            for x in get_commit_logs(
+                tag=tag,
+                all_logs=all_logs,
+                is_dt=datetime_mode,
+            )
+        ),
         file=sys.stdout,
     )
     sys.exit(0)
 
 
 @cli_git.command()
-@click.option("-f", "--file", type=click.STRING, default=None)
-@click.option("-l", "--latest", is_flag=True)
+@click.argument(
+    "file",
+    type=click.STRING,
+    default=".git/COMMIT_EDITMSG",
+)
 @click.option("-e", "--edit", is_flag=True)
 @click.option("-o", "--output-file", is_flag=True)
 @click.option("-p", "--prepare", is_flag=True)
 def cm(
     file: Optional[str],
-    latest: bool,
     edit: bool,
     output_file: bool,
     prepare: bool,
-):
+):  # pragma: no cover.
     """Show the latest Commit message"""
-    if latest and not file:
-        file = ".git/COMMIT_EDITMSG"
     if not prepare:
         click.echo(
             make_color(
@@ -411,7 +510,7 @@ def cm(
         )
     else:
         edit: bool = True
-        cm_msg: str = "\n".join(get_latest_commit(file, edit, output_file))
+        _cm_msg: str = "\n".join(get_latest_commit(file, edit, output_file))
         subprocess.run(
             [
                 "git",
@@ -420,35 +519,111 @@ def cm(
                 "-a",
                 "--no-verify",
                 "-m",
-                cm_msg,
+                _cm_msg,
             ],
             stdout=subprocess.DEVNULL,
         )
-        click.echo(make_color(cm_msg, level=Level.OK))
+        click.echo(make_color(_cm_msg, level=Level.OK))
     sys.exit(0)
 
 
 @cli_git.command()
-@click.option("--no-verify", is_flag=True)
-def cm_previous(no_verify: bool):
+@click.option("-g", "--group", is_flag=True)
+def cm_msg(group: bool = False) -> NoReturn:  # pragma: no cover.
+    """Return list of commit prefixes"""
+    if group:
+        for cm_prefix_g in get_commit_prefix_group():
+            click.echo(f"{cm_prefix_g.emoji} {cm_prefix_g.name}")
+    else:
+        for cm_prefix in get_commit_prefix():
+            click.echo(
+                f"{cm_prefix.emoji} {cm_prefix.name} -> {cm_prefix.group}"
+            )
+    sys.exit(0)
+
+
+@cli_git.command()
+@click.option("--verify", is_flag=True)
+def cm_prev(verify: bool) -> NoReturn:  # pragma: no cover.
     """Commit changes to the Previous Commit with same message."""
-    merge2latest_commit(no_verify=no_verify)
+    merge2latest_commit(no_verify=(not verify))
     sys.exit(0)
 
 
 @cli_git.command()
 @click.option("-f", "--force", is_flag=True)
 @click.option("-n", "--number", type=click.INT, default=1)
-def cm_revert(force: bool, number: int):
+def cm_revert(force: bool, number: int):  # pragma: no cover.
     """Revert the latest Commit on the Local repository."""
     subprocess.run(["git", "reset", f"HEAD~{number}"])
     if force:
         subprocess.run(["git", "restore", "."])
+        subprocess.run(["git", "clean", "-f"])
     sys.exit(0)
 
 
 @cli_git.command()
-def bn_clear():
+@click.argument(
+    "branch",
+    type=click.STRING,
+    default="main",
+)
+@click.option(
+    "-t",
+    "--theirs",
+    is_flag=True,
+    help="If True, it will use `their` strategy if it has conflict",
+)
+@click.option(
+    "-o",
+    "--ours",
+    is_flag=True,
+    help="If True, it will use `ours` strategy if it has conflict",
+)
+@click.option(
+    "-s",
+    "--squash",
+    is_flag=True,
+    help="If True, it will use `squash` merge option.",
+)
+def mg(
+    branch: str,
+    theirs: bool = False,
+    ours: bool = False,
+    squash: bool = False,
+) -> NoReturn:  # pragma: no cover.
+    """Merge change from another branch with strategy, `theirs` or `ours`.
+
+    BRANCH is a name of branch that you want to merge with current branch.
+
+    \f
+    :param branch:
+    :param theirs:
+    :param ours:
+    :param squash:
+    """
+    if theirs and ours:
+        raise ValueError("The strategy flag should not True together.")
+    elif ours:
+        strategy = "ours"
+    else:
+        strategy = "theirs"
+    subprocess.run(
+        [
+            "git",
+            "merge",
+            branch,
+            "--strategy-option",
+            strategy,
+            *(["--squash"] if squash else []),
+        ],
+        stderr=subprocess.DEVNULL,
+    )
+    sys.exit(0)
+
+
+@cli_git.command()
+def bn_clear() -> NoReturn:  # pragma: no cover.
     """Clear Local Branches that sync from the Remote repository."""
     subprocess.run(
         ["git", "checkout", "main"],
@@ -474,7 +649,31 @@ def bn_clear():
 
 
 @cli_git.command()
-def tg_clear():
+@click.option(
+    "-p",
+    "--push",
+    is_flag=True,
+    help="If True, it will auto push to remote",
+)
+def tg_bump(push: bool = False) -> NoReturn:  # pragma: no cover.
+    """Create Tag from current version after bumping"""
+    latest_tag: str = get_latest_tag(default=False)
+    subprocess.run(
+        ["git", "tag", "-d", f"v{latest_tag}"],
+        stderr=subprocess.DEVNULL,
+    )
+    subprocess.run(
+        ["git", "fetch", "--prune", "--prune-tags"],
+        stdout=subprocess.DEVNULL,
+    )
+    subprocess.run(["git", "tag", f"v{latest_tag}"])
+    if push:
+        subprocess.run(["git", "push", f"v{latest_tag}", "--tags"])
+    sys.exit(0)
+
+
+@cli_git.command()
+def tg_clear() -> NoReturn:  # pragma: no cover.
     """Clear Local Tags that sync from the Remote repository."""
     subprocess.run(
         ["git", "fetch", "--prune", "--prune-tags"],
@@ -494,7 +693,7 @@ def tg_clear():
     is_flag=True,
     help="If True, it will set fetch handle prune tag.",
 )
-def init(store: bool, prune_tag: bool):
+def init(store: bool, prune_tag: bool) -> NoReturn:  # pragma: no cover.
     """Initialize GIT config on local"""
     if not Path(".git").exists():
         click.echo("Start initialize git on current path ...")
@@ -525,11 +724,19 @@ def init(store: bool, prune_tag: bool):
 
 
 @cli_git.command()
-def pf():
+def pf() -> NoReturn:  # pragma: no cover.
     """Show Profile object that contain Name and Email of Author"""
     click.echo(load_profile(), file=sys.stdout)
     sys.exit(0)
 
 
+@cli_git.command()
+def df() -> NoReturn:  # pragma: no cover.
+    """Show changed files from previous commit to HEAD"""
+    # NOTE: We can use `git show --name-only HEAD~1`, but it got commit message.
+    subprocess.run(["git", "diff", "--name-only", "HEAD~1", "HEAD"])
+    sys.exit(0)
+
+
 if __name__ == "__main__":
     cli_git.main()
